@@ -1,7 +1,7 @@
 //! Bounds-checking for SPIR-V output.
 
-use super::{Block, BlockContext, Error, IdGenerator, Instruction, Word};
-use crate::{arena::Handle, back::IndexBoundsCheckPolicy};
+use super::{selection::Selection, Block, BlockContext, Error, IdGenerator, Instruction, Word};
+use crate::{arena::Handle, back::BoundsCheckPolicy};
 
 /// The results of emitting code for a left-hand-side expression.
 ///
@@ -25,7 +25,7 @@ pub(super) enum ExpressionPointer {
 /// The results of performing a bounds check.
 ///
 /// On success, `write_bounds_check` returns a value of this type.
-pub enum BoundsCheckResult {
+pub(super) enum BoundsCheckResult {
     /// The index is statically known and in bounds, with the given value.
     KnownInBounds(u32),
 
@@ -38,7 +38,7 @@ pub enum BoundsCheckResult {
 }
 
 /// A value that we either know at translation time, or need to compute at runtime.
-pub enum MaybeKnown<T> {
+pub(super) enum MaybeKnown<T> {
     /// The value is known at shader translation time.
     Known(T),
 
@@ -141,7 +141,7 @@ impl<'w> BlockContext<'w> {
             }
             MaybeKnown::Computed(length_id) => {
                 // Emit code to compute the max index from the length.
-                let const_one_id = self.get_index_constant(1)?;
+                let const_one_id = self.get_index_constant(1);
                 let max_index_id = self.gen_id();
                 block.body.push(Instruction::binary(
                     spirv::Op::ISub,
@@ -157,11 +157,11 @@ impl<'w> BlockContext<'w> {
 
     /// Restrict an index to be in range for a vector, matrix, or array.
     ///
-    /// This is used to implement `IndexBoundsCheckPolicy::Restrict`. An
-    /// in-bounds index is left unchanged. An out-of-bounds index is replaced
-    /// with some arbitrary in-bounds index. Note,this is not necessarily
-    /// clamping; for example, negative indices might be changed to refer to the
-    /// last element of the sequence, not the first, as clamping would do.
+    /// This is used to implement `BoundsCheckPolicy::Restrict`. An in-bounds
+    /// index is left unchanged. An out-of-bounds index is replaced with some
+    /// arbitrary in-bounds index. Note,this is not necessarily clamping; for
+    /// example, negative indices might be changed to refer to the last element
+    /// of the sequence, not the first, as clamping would do.
     ///
     /// Either return the restricted index value, if known, or add instructions
     /// to `block` to compute it, and return the id of the result. See the
@@ -196,13 +196,13 @@ impl<'w> BlockContext<'w> {
                     }
                 }
 
-                self.get_index_constant(known_max_index)?
+                self.get_index_constant(known_max_index)
             }
             MaybeKnown::Computed(max_index_id) => max_index_id,
         };
 
         // One or the other of the index or length is dynamic, so emit code for
-        // IndexBoundsCheckPolicy::Restrict.
+        // BoundsCheckPolicy::Restrict.
         let restricted_index_id = self.gen_id();
         block.body.push(Instruction::ext_inst(
             self.writer.gl450_ext_inst_id,
@@ -271,7 +271,7 @@ impl<'w> BlockContext<'w> {
                     }
                 }
 
-                self.get_index_constant(known_length)?
+                self.get_index_constant(known_length)
             }
             MaybeKnown::Computed(length_id) => length_id,
         };
@@ -290,7 +290,7 @@ impl<'w> BlockContext<'w> {
         Ok(BoundsCheckResult::Conditional(condition_id))
     }
 
-    /// Emit a conditional load for `IndexBoundsCheckPolicy::ReadZeroSkipWrite`.
+    /// Emit a conditional load for `BoundsCheckPolicy::ReadZeroSkipWrite`.
     ///
     /// Generate code to load a value of `result_type` if `condition` is true,
     /// and generate a null value of that type if it is false. Call `emit_load`
@@ -306,52 +306,37 @@ impl<'w> BlockContext<'w> {
     where
         F: FnOnce(&mut IdGenerator, &mut Block) -> Word,
     {
-        let header_block = block.label_id;
-        let merge_block = self.gen_id();
-        let in_bounds_block = self.gen_id();
-
-        // Branch based on whether the index was in bounds.
-        //
-        // As it turns out, our out-of-bounds branch block would contain no
-        // instructions: it just produces a constant zero, whose instruction is
-        // in the module's declarations section at the front. In this case,
-        // SPIR-V lets us omit the empty 'else' block, and branch directly to
-        // the merge block. The phi instruction in the merge block can cite the
-        // header block as its CFG predecessor.
-        block.body.push(Instruction::selection_merge(
-            merge_block,
-            spirv::SelectionControl::NONE,
-        ));
-        self.function.consume(
-            std::mem::replace(block, Block::new(in_bounds_block)),
-            Instruction::branch_conditional(condition, in_bounds_block, merge_block),
-        );
-
-        // The in-bounds path. Perform the access and the load.
-        let value_id = emit_load(&mut self.writer.id_gen, block);
-
-        // Finish the in-bounds block and start the merge block. This
-        // is the block we'll leave current on return.
-        self.function.consume(
-            std::mem::replace(block, Block::new(merge_block)),
-            Instruction::branch(merge_block),
-        );
-
-        // For the out-of-bounds case, produce a zero value.
+        // For the out-of-bounds case, we produce a zero value.
         let null_id = self.writer.write_constant_null(result_type);
 
-        // Merge the results from the two paths.
-        let result_id = self.gen_id();
-        block.body.push(Instruction::phi(
-            result_type,
-            result_id,
-            &[(value_id, in_bounds_block), (null_id, header_block)],
-        ));
+        let mut selection = Selection::start(block, result_type);
 
-        result_id
+        // As it turns out, we don't actually need a full 'if-then-else'
+        // structure for this: SPIR-V constants are declared up front, so the
+        // 'else' block would have no instructions. Instead we emit something
+        // like this:
+        //
+        //     result = zero;
+        //     if in_bounds {
+        //         result = do the load;
+        //     }
+        //     use result;
+
+        // Continue only if the index was in bounds. Otherwise, branch to the
+        // merge block.
+        selection.if_true(self, condition, null_id);
+
+        // The in-bounds path. Perform the access and the load.
+        let loaded_value = emit_load(&mut self.writer.id_gen, selection.block());
+
+        selection.finish(self, loaded_value)
     }
 
-    /// Emit code for bounds checks, per self.index_bounds_check_policy.
+    /// Emit code for bounds checks for an array, vector, or matrix access.
+    ///
+    /// This implements either `index_bounds_check_policy` or
+    /// `buffer_bounds_check_policy`, depending on the storage class of the
+    /// pointer being accessed.
     ///
     /// Return a `BoundsCheckResult` indicating how the index should be
     /// consumed. See that type's documentation for details.
@@ -361,14 +346,29 @@ impl<'w> BlockContext<'w> {
         index: Handle<crate::Expression>,
         block: &mut Block,
     ) -> Result<BoundsCheckResult, Error> {
-        Ok(match self.writer.index_bounds_check_policy {
-            IndexBoundsCheckPolicy::Restrict => self.write_restricted_index(base, index, block)?,
-            IndexBoundsCheckPolicy::ReadZeroSkipWrite => {
+        // Should this access be covered by `index_bounds_check_policy` or
+        // `buffer_bounds_check_policy`?
+        let is_buffer = match *self.fun_info[base].ty.inner_with(&self.ir_module.types) {
+            crate::TypeInner::Pointer { class, .. }
+            | crate::TypeInner::ValuePointer { class, .. } => match class {
+                crate::StorageClass::Storage { access: _ } | crate::StorageClass::Uniform => true,
+                _ => false,
+            },
+            _ => false,
+        };
+
+        let policy = if is_buffer {
+            self.writer.bounds_check_policies.buffer
+        } else {
+            self.writer.bounds_check_policies.index
+        };
+
+        Ok(match policy {
+            BoundsCheckPolicy::Restrict => self.write_restricted_index(base, index, block)?,
+            BoundsCheckPolicy::ReadZeroSkipWrite => {
                 self.write_index_comparison(base, index, block)?
             }
-            IndexBoundsCheckPolicy::UndefinedBehavior => {
-                BoundsCheckResult::Computed(self.cached[index])
-            }
+            BoundsCheckPolicy::Unchecked => BoundsCheckResult::Computed(self.cached[index]),
         })
     }
 
