@@ -3,15 +3,17 @@
 
 mod block;
 mod helpers;
+mod image;
 mod index;
 mod instructions;
 mod layout;
 mod recyclable;
+mod selection;
 mod writer;
 
 pub use spirv::Capability;
 
-use crate::{arena::Handle, back::IndexBoundsCheckPolicy, proc::TypeResolution};
+use crate::{arena::Handle, back::BoundsCheckPolicies, proc::TypeResolution};
 
 use spirv::Word;
 use std::ops;
@@ -53,6 +55,8 @@ const BITS_PER_BYTE: crate::Bytes = 8;
 
 #[derive(Clone, Debug, Error)]
 pub enum Error {
+    #[error("The requested entry point couldn't be found")]
+    EntryPointNotFound,
     #[error("target SPIRV-{0}.{1} is not supported")]
     UnsupportedVersion(u8, u8),
     #[error("using {0} requires at least one of the capabilities {1:?}, but none are available")]
@@ -214,19 +218,47 @@ impl LocalImageType {
 
 /// A SPIR-V type constructed during code generation.
 ///
-/// In the process of writing SPIR-V, we need to synthesize various types for
-/// intermediate results and such. However, it's inconvenient to use
-/// `crate::Type` or `crate::TypeInner` for these, as the IR module is immutable
-/// so we can't ever create a `Handle<Type>` to refer to them. So for local use
-/// in the SPIR-V writer, we have this home-grown type enum that covers only the
-/// cases we need (for example, it doesn't cover structs).
+/// This is the variant of [`LookupType`] used to represent types that might not
+/// be available in the arena. Variants are present here for one of two reasons:
 ///
-/// As explained in §2.8 of the SPIR-V spec, some classes of type instructions
-/// must be unique; for example, you can't have two `OpTypeInt 32 1`
-/// instructions in the same module. `Writer::lookup_type` maps each `LocalType`
-/// value for which we've written instructions to its id, so we can avoid
-/// writing out duplicates. `LocalType` also includes variants like `Pointer`
-/// that do not need to be unique - but it is harmless to avoid the duplication.
+/// -   They represent types synthesized during code generation, as explained
+///     in the documentation for [`LookupType`].
+///
+/// -   They represent types for which SPIR-V forbids duplicate `OpType...`
+///     instructions, requiring deduplication.
+///
+/// This is not a complete copy of [`TypeInner`]: for example, SPIR-V generation
+/// never synthesizes new struct types, so `LocalType` has nothing for that.
+///
+/// Each `LocalType` variant should be handled identically to its analogous
+/// `TypeInner` variant. You can use the [`make_local`] function to help with
+/// this, by converting everything possible to a `LocalType` before inspecting
+/// it.
+///
+/// ## `Localtype` equality and SPIR-V `OpType` uniqueness
+///
+/// The definition of `Eq` on `LocalType` is carefully chosen to help us follow
+/// certain SPIR-V rules. SPIR-V §2.8 requires some classes of `OpType...`
+/// instructions to be unique; for example, you can't have two `OpTypeInt 32 1`
+/// instructions in the same module. All 32-bit signed integers must use the
+/// same type id.
+///
+/// All SPIR-V types that must be unique can be represented as a `LocalType`,
+/// and two `LocalType`s are always `Eq` if SPIR-V would require them to use the
+/// same `OpType...` instruction. This lets us avoid duplicates by recording the
+/// ids of the type instructions we've already generated in a hash table,
+/// [`Writer::lookup_type`], keyed by `LocalType`.
+///
+/// As another example, [`LocalImageType`], stored in the `LocalType::Image`
+/// variant, is designed to help us deduplicate `OpTypeImage` instructions. See
+/// its documentation for details.
+///
+/// `LocalType` also includes variants like `Pointer` that do not need to be
+/// unique - but it is harmless to avoid the duplication.
+///
+/// As it always must, the `Hash` implementation respects the `Eq` relation.
+///
+/// [`TypeInner`]: crate::TypeInner
 #[derive(Debug, PartialEq, Hash, Eq, Copy, Clone)]
 enum LocalType {
     /// A scalar, vector, or pointer to one of those.
@@ -255,6 +287,28 @@ enum LocalType {
     Sampler,
 }
 
+/// A type encountered during SPIR-V generation.
+///
+/// In the process of writing SPIR-V, we need to synthesize various types for
+/// intermediate results and such: pointer types, vector/matrix component types,
+/// or even booleans, which usually appear in SPIR-V code even when they're not
+/// used by the module source.
+///
+/// However, we can't use `crate::Type` or `crate::TypeInner` for these, as the
+/// type arena may not contain what we need (it only contains types used
+/// directly by other parts of the IR), and the IR module is immutable, so we
+/// can't add anything to it.
+///
+/// So for local use in the SPIR-V writer, we use this type, which holds either
+/// a handle into the arena, or a [`LocalType`] containing something synthesized
+/// locally.
+///
+/// This is very similar to the [`proc::TypeResolution`] enum, with `LocalType`
+/// playing the role of `TypeInner`. However, `LocalType` also has other
+/// properties needed for SPIR-V generation; see the description of
+/// [`LocalType`] for details.
+///
+/// [`proc::TypeResolution`]: crate::proc::TypeResolution
 #[derive(Debug, PartialEq, Hash, Eq, Copy, Clone)]
 enum LookupType {
     Handle(Handle<crate::Type>),
@@ -384,6 +438,24 @@ struct GlobalVariable {
     handle_id: Word,
 }
 
+impl GlobalVariable {
+    fn dummy() -> Self {
+        Self {
+            id: 0,
+            handle_id: 0,
+        }
+    }
+
+    fn new(id: Word) -> Self {
+        Self { id, handle_id: 0 }
+    }
+
+    /// Prepare `self` for use within a single function.
+    fn reset_for_function(&mut self) {
+        self.handle_id = 0;
+    }
+}
+
 struct FunctionArgument {
     /// Actual instruction of the argument.
     instruction: Instruction,
@@ -420,16 +492,20 @@ impl BlockContext<'_> {
         self.writer.id_gen.next()
     }
 
+    fn get_type_id(&mut self, lookup_type: LookupType) -> Word {
+        self.writer.get_type_id(lookup_type)
+    }
+
     fn get_expression_type_id(&mut self, tr: &TypeResolution) -> Word {
         self.writer.get_expression_type_id(tr)
     }
 
-    fn get_index_constant(&mut self, index: Word) -> Result<Word, Error> {
+    fn get_index_constant(&mut self, index: Word) -> Word {
         self.writer
             .get_constant_scalar(crate::ScalarValue::Uint(index as _), 4)
     }
 
-    fn get_scope_constant(&mut self, scope: Word) -> Result<Word, Error> {
+    fn get_scope_constant(&mut self, scope: Word) -> Word {
         self.writer
             .get_constant_scalar(crate::ScalarValue::Sint(scope as _), 4)
     }
@@ -460,7 +536,7 @@ pub struct Writer {
     debugs: Vec<Instruction>,
     annotations: Vec<Instruction>,
     flags: WriterFlags,
-    index_bounds_check_policy: IndexBoundsCheckPolicy,
+    bounds_check_policies: BoundsCheckPolicies,
     void_type: Word,
     //TODO: convert most of these into vectors, addressable by handle indices
     lookup_type: crate::FastHashMap<LookupType, Word>,
@@ -485,9 +561,15 @@ bitflags::bitflags! {
         const DEBUG = 0x1;
         /// Flip Y coordinate of `BuiltIn::Position` output.
         const ADJUST_COORDINATE_SPACE = 0x2;
-        /// Emit `OpLabel` for input/output locations.
-        /// Some drivers treat it as semantic, not allowing any conflicts.
+        /// Emit `OpName` for input/output locations.
+        /// Contrary to spec, some drivers treat it as semantic, not allowing
+        /// any conflicts.
         const LABEL_VARYINGS = 0x4;
+        /// Emit `PointSize` output builtin to vertex shaders, which is
+        /// required for drawing with `PointList` topology.
+        const FORCE_POINT_SIZE = 0x8;
+        /// Clamp `BuiltIn::FragDepth` output between 0 and 1.
+        const CLAMP_FRAG_DEPTH = 0x10;
     }
 }
 
@@ -505,14 +587,16 @@ pub struct Options {
     /// If this is `None`, all capabilities are permitted.
     pub capabilities: Option<crate::FastHashSet<Capability>>,
 
-    /// How should the generated code handle array, vector, or matrix indices
-    /// that are out of range?
-    pub index_bounds_check_policy: IndexBoundsCheckPolicy,
+    /// How should generate code handle array, vector, matrix, or image texel
+    /// indices that are out of range?
+    pub bounds_check_policies: BoundsCheckPolicies,
 }
 
 impl Default for Options {
     fn default() -> Self {
-        let mut flags = WriterFlags::ADJUST_COORDINATE_SPACE | WriterFlags::LABEL_VARYINGS;
+        let mut flags = WriterFlags::ADJUST_COORDINATE_SPACE
+            | WriterFlags::LABEL_VARYINGS
+            | WriterFlags::CLAMP_FRAG_DEPTH;
         if cfg!(debug_assertions) {
             flags |= WriterFlags::DEBUG;
         }
@@ -520,18 +604,33 @@ impl Default for Options {
             lang_version: (1, 0),
             flags,
             capabilities: None,
-            index_bounds_check_policy: super::IndexBoundsCheckPolicy::default(),
+            bounds_check_policies: super::BoundsCheckPolicies::default(),
         }
     }
 }
 
+// A subset of options that are meant to be changed per pipeline.
+#[derive(Debug, Clone, PartialEq, Eq, Hash)]
+#[cfg_attr(feature = "serialize", derive(serde::Serialize))]
+#[cfg_attr(feature = "deserialize", derive(serde::Deserialize))]
+pub struct PipelineOptions {
+    /// The stage of the entry point
+    pub shader_stage: crate::ShaderStage,
+    /// The name of the entry point
+    ///
+    /// If no entry point that matches is found a error will be thrown while creating a new instance
+    /// of [`Writer`](struct.Writer.html)
+    pub entry_point: String,
+}
+
 pub fn write_vec(
     module: &crate::Module,
     info: &crate::valid::ModuleInfo,
     options: &Options,
+    pipeline_options: Option<&PipelineOptions>,
 ) -> Result<Vec<u32>, Error> {
     let mut words = Vec::new();
     let mut w = Writer::new(options)?;
-    w.write(module, info, &mut words)?;
+    w.write(module, info, pipeline_options, &mut words)?;
     Ok(words)
 }
