@@ -1,4 +1,5 @@
-use crate::front::glsl::SourceMetadata;
+use crate::front::glsl::context::ExprPos;
+use crate::front::glsl::Span;
 use crate::{
     front::glsl::{
         ast::ParameterQualifier,
@@ -39,7 +40,8 @@ impl<'source> ParsingContext<'source> {
         parser: &mut Parser,
         ctx: &mut Context,
         body: &mut Block,
-    ) -> Result<Option<SourceMetadata>> {
+        terminator: &mut Option<usize>,
+    ) -> Result<Option<Span>> {
         // TODO: This prevents snippets like the following from working
         // ```glsl
         // vec4(1.0);
@@ -53,19 +55,26 @@ impl<'source> ParsingContext<'source> {
 
         let new_break = || {
             let mut block = Block::new();
-            block.push(Statement::Break, crate::Span::Unknown);
+            block.push(Statement::Break, crate::Span::default());
             block
         };
 
-        let &Token { ref value, meta } = self.expect_peek(parser)?;
+        let &Token {
+            ref value,
+            mut meta,
+        } = self.expect_peek(parser)?;
 
         let meta_rest = match *value {
             TokenValue::Continue => {
-                body.push(Statement::Continue, self.bump(parser)?.meta.as_span());
+                let meta = self.bump(parser)?.meta;
+                body.push(Statement::Continue, meta);
+                terminator.get_or_insert(body.len());
                 self.expect(parser, TokenValue::Semicolon)?.meta
             }
             TokenValue::Break => {
-                body.push(Statement::Break, self.bump(parser)?.meta.as_span());
+                let meta = self.bump(parser)?.meta;
+                body.push(Statement::Break, meta);
+                terminator.get_or_insert(body.len());
                 self.expect(parser, TokenValue::Semicolon)?.meta
             }
             TokenValue::Return => {
@@ -77,7 +86,8 @@ impl<'source> ParsingContext<'source> {
                         let mut stmt = ctx.stmt_ctx();
                         let expr = self.parse_expression(parser, ctx, &mut stmt, body)?;
                         self.expect(parser, TokenValue::Semicolon)?;
-                        let (handle, meta) = ctx.lower_expect(stmt, parser, expr, false, body)?;
+                        let (handle, meta) =
+                            ctx.lower_expect(stmt, parser, expr, ExprPos::Rhs, body)?;
                         (Some(handle), meta)
                     }
                 };
@@ -85,11 +95,16 @@ impl<'source> ParsingContext<'source> {
                 ctx.emit_flush(body);
                 ctx.emit_start();
 
-                body.push(Statement::Return { value }, meta.as_span());
+                body.push(Statement::Return { value }, meta);
+                terminator.get_or_insert(body.len());
+
                 meta
             }
             TokenValue::Discard => {
-                body.push(Statement::Kill, self.bump(parser)?.meta.as_span());
+                let meta = self.bump(parser)?.meta;
+                body.push(Statement::Kill, meta);
+                terminator.get_or_insert(body.len());
+
                 self.expect(parser, TokenValue::Semicolon)?.meta
             }
             TokenValue::If => {
@@ -99,8 +114,9 @@ impl<'source> ParsingContext<'source> {
                 let condition = {
                     let mut stmt = ctx.stmt_ctx();
                     let expr = self.parse_expression(parser, ctx, &mut stmt, body)?;
-                    let (handle, more_meta) = ctx.lower_expect(stmt, parser, expr, false, body)?;
-                    meta = meta.union(&more_meta);
+                    let (handle, more_meta) =
+                        ctx.lower_expect(stmt, parser, expr, ExprPos::Rhs, body)?;
+                    meta.subsume(more_meta);
                     handle
                 };
                 self.expect(parser, TokenValue::RightParen)?;
@@ -109,14 +125,18 @@ impl<'source> ParsingContext<'source> {
                 ctx.emit_start();
 
                 let mut accept = Block::new();
-                if let Some(more_meta) = self.parse_statement(parser, ctx, &mut accept)? {
-                    meta = meta.union(&more_meta)
+                if let Some(more_meta) =
+                    self.parse_statement(parser, ctx, &mut accept, &mut None)?
+                {
+                    meta.subsume(more_meta)
                 }
 
                 let mut reject = Block::new();
                 if self.bump_if(parser, TokenValue::Else).is_some() {
-                    if let Some(more_meta) = self.parse_statement(parser, ctx, &mut reject)? {
-                        meta = meta.union(&more_meta);
+                    if let Some(more_meta) =
+                        self.parse_statement(parser, ctx, &mut reject, &mut None)?
+                    {
+                        meta.subsume(more_meta);
                     }
                 }
 
@@ -126,21 +146,23 @@ impl<'source> ParsingContext<'source> {
                         accept,
                         reject,
                     },
-                    meta.as_span(),
+                    meta,
                 );
+
                 meta
             }
             TokenValue::Switch => {
-                let start_meta = self.bump(parser)?.meta;
+                let mut meta = self.bump(parser)?.meta;
                 let end_meta;
 
                 self.expect(parser, TokenValue::LeftParen)?;
-                // TODO: Implicit conversions
+
                 let selector = {
                     let mut stmt = ctx.stmt_ctx();
                     let expr = self.parse_expression(parser, ctx, &mut stmt, body)?;
-                    ctx.lower_expect(stmt, parser, expr, false, body)?.0
+                    ctx.lower_expect(stmt, parser, expr, ExprPos::Rhs, body)?.0
                 };
+
                 self.expect(parser, TokenValue::RightParen)?;
 
                 ctx.emit_flush(body);
@@ -158,7 +180,7 @@ impl<'source> ParsingContext<'source> {
                                 let mut stmt = ctx.stmt_ctx();
                                 let expr = self.parse_expression(parser, ctx, &mut stmt, body)?;
                                 let (root, meta) =
-                                    ctx.lower_expect(stmt, parser, expr, false, body)?;
+                                    ctx.lower_expect(stmt, parser, expr, ExprPos::Rhs, body)?;
                                 let constant = parser.solve_constant(ctx, root, meta)?;
 
                                 match parser.module.constants[constant].inner {
@@ -187,25 +209,32 @@ impl<'source> ParsingContext<'source> {
 
                             let mut body = Block::new();
 
+                            let mut case_terminator = None;
                             loop {
                                 match self.expect_peek(parser)?.value {
                                     TokenValue::Case
                                     | TokenValue::Default
                                     | TokenValue::RightBrace => break,
                                     _ => {
-                                        self.parse_statement(parser, ctx, &mut body)?;
+                                        self.parse_statement(
+                                            parser,
+                                            ctx,
+                                            &mut body,
+                                            &mut case_terminator,
+                                        )?;
                                     }
                                 }
                             }
 
                             let mut fall_through = true;
 
-                            for (i, stmt) in body.iter().enumerate() {
-                                if let Statement::Break = *stmt {
+                            if let Some(mut idx) = case_terminator {
+                                if let Statement::Break = body[idx - 1] {
                                     fall_through = false;
-                                    body.cull(i..);
-                                    break;
+                                    idx -= 1;
                                 }
+
+                                body.cull(idx..)
                             }
 
                             cases.push(SwitchCase {
@@ -228,21 +257,20 @@ impl<'source> ParsingContext<'source> {
                                 });
                             }
 
+                            let mut default_terminator = None;
                             loop {
                                 match self.expect_peek(parser)?.value {
                                     TokenValue::Case | TokenValue::RightBrace => break,
                                     _ => {
-                                        self.parse_statement(parser, ctx, &mut default)?;
+                                        self.parse_statement(
+                                            parser,
+                                            ctx,
+                                            &mut default,
+                                            &mut default_terminator,
+                                        )?;
                                     }
                                 }
                             }
-
-                            for (i, stmt) in default.iter().enumerate() {
-                                if let Statement::Break = *stmt {
-                                    default.cull(i..);
-                                    break;
-                                }
-                            }
                         }
                         TokenValue::RightBrace => {
                             end_meta = self.bump(parser)?.meta;
@@ -265,29 +293,31 @@ impl<'source> ParsingContext<'source> {
                     }
                 }
 
-                let meta = start_meta.union(&end_meta);
+                meta.subsume(end_meta);
+
                 body.push(
                     Statement::Switch {
                         selector,
                         cases,
                         default,
                     },
-                    meta.as_span(),
+                    meta,
                 );
+
                 meta
             }
             TokenValue::While => {
-                let meta = self.bump(parser)?.meta;
+                let mut meta = self.bump(parser)?.meta;
 
                 let mut loop_body = Block::new();
 
                 let mut stmt = ctx.stmt_ctx();
                 self.expect(parser, TokenValue::LeftParen)?;
                 let root = self.parse_expression(parser, ctx, &mut stmt, &mut loop_body)?;
-                let meta = meta.union(&self.expect(parser, TokenValue::RightParen)?.meta);
+                meta.subsume(self.expect(parser, TokenValue::RightParen)?.meta);
 
                 let (expr, expr_meta) =
-                    ctx.lower_expect(stmt, parser, root, false, &mut loop_body)?;
+                    ctx.lower_expect(stmt, parser, root, ExprPos::Rhs, &mut loop_body)?;
                 let condition = ctx.add_expression(
                     Expression::Unary {
                         op: UnaryOperator::Not,
@@ -306,13 +336,15 @@ impl<'source> ParsingContext<'source> {
                         accept: new_break(),
                         reject: Block::new(),
                     },
-                    crate::Span::Unknown,
+                    crate::Span::default(),
                 );
 
-                let mut meta = meta.union(&expr_meta);
+                meta.subsume(expr_meta);
 
-                if let Some(body_meta) = self.parse_statement(parser, ctx, &mut loop_body)? {
-                    meta = meta.union(&body_meta);
+                if let Some(body_meta) =
+                    self.parse_statement(parser, ctx, &mut loop_body, &mut None)?
+                {
+                    meta.subsume(body_meta);
                 }
 
                 body.push(
@@ -320,15 +352,17 @@ impl<'source> ParsingContext<'source> {
                         body: loop_body,
                         continuing: Block::new(),
                     },
-                    meta.as_span(),
+                    meta,
                 );
+
                 meta
             }
             TokenValue::Do => {
-                let start_meta = self.bump(parser)?.meta;
+                let mut meta = self.bump(parser)?.meta;
 
                 let mut loop_body = Block::new();
-                self.parse_statement(parser, ctx, &mut loop_body)?;
+
+                self.parse_statement(parser, ctx, &mut loop_body, &mut None)?;
 
                 let mut stmt = ctx.stmt_ctx();
 
@@ -337,10 +371,10 @@ impl<'source> ParsingContext<'source> {
                 let root = self.parse_expression(parser, ctx, &mut stmt, &mut loop_body)?;
                 let end_meta = self.expect(parser, TokenValue::RightParen)?.meta;
 
-                let meta = start_meta.union(&end_meta);
+                meta.subsume(end_meta);
 
                 let (expr, expr_meta) =
-                    ctx.lower_expect(stmt, parser, root, false, &mut loop_body)?;
+                    ctx.lower_expect(stmt, parser, root, ExprPos::Rhs, &mut loop_body)?;
                 let condition = ctx.add_expression(
                     Expression::Unary {
                         op: UnaryOperator::Not,
@@ -359,7 +393,7 @@ impl<'source> ParsingContext<'source> {
                         accept: new_break(),
                         reject: Block::new(),
                     },
-                    crate::Span::Unknown,
+                    crate::Span::default(),
                 );
 
                 body.push(
@@ -367,12 +401,13 @@ impl<'source> ParsingContext<'source> {
                         body: loop_body,
                         continuing: Block::new(),
                     },
-                    meta.as_span(),
+                    meta,
                 );
+
                 meta
             }
             TokenValue::For => {
-                let meta = self.bump(parser)?.meta;
+                let mut meta = self.bump(parser)?.meta;
 
                 ctx.push_scope();
                 self.expect(parser, TokenValue::LeftParen)?;
@@ -383,7 +418,7 @@ impl<'source> ParsingContext<'source> {
                     } else {
                         let mut stmt = ctx.stmt_ctx();
                         let expr = self.parse_expression(parser, ctx, &mut stmt, body)?;
-                        ctx.lower(stmt, parser, expr, false, body)?;
+                        ctx.lower(stmt, parser, expr, ExprPos::Rhs, body)?;
                         self.expect(parser, TokenValue::Semicolon)?;
                     }
                 }
@@ -394,14 +429,14 @@ impl<'source> ParsingContext<'source> {
                     let (expr, expr_meta) =
                         if self.peek_type_name(parser) || self.peek_type_qualifier(parser) {
                             let qualifiers = self.parse_type_qualifiers(parser)?;
-                            let (ty, meta) = self.parse_type_non_void(parser)?;
+                            let (ty, mut meta) = self.parse_type_non_void(parser)?;
                             let name = self.expect_ident(parser)?.0;
 
                             self.expect(parser, TokenValue::Assign)?;
 
                             let (value, end_meta) =
                                 self.parse_initializer(parser, ty, ctx, &mut block)?;
-                            let meta = meta.union(&end_meta);
+                            meta.subsume(end_meta);
 
                             let decl = VarDeclaration {
                                 qualifiers: &qualifiers,
@@ -416,13 +451,13 @@ impl<'source> ParsingContext<'source> {
                             ctx.emit_flush(&mut block);
                             ctx.emit_start();
 
-                            block.push(Statement::Store { pointer, value }, meta.as_span());
+                            block.push(Statement::Store { pointer, value }, meta);
 
                             (value, end_meta)
                         } else {
                             let mut stmt = ctx.stmt_ctx();
                             let root = self.parse_expression(parser, ctx, &mut stmt, &mut block)?;
-                            ctx.lower_expect(stmt, parser, root, false, &mut block)?
+                            ctx.lower_expect(stmt, parser, root, ExprPos::Rhs, &mut block)?
                         };
 
                     let condition = ctx.add_expression(
@@ -443,7 +478,7 @@ impl<'source> ParsingContext<'source> {
                             accept: new_break(),
                             reject: Block::new(),
                         },
-                        crate::Span::Unknown,
+                        crate::Span::default(),
                     );
 
                     self.expect(parser, TokenValue::Semicolon)?;
@@ -455,14 +490,14 @@ impl<'source> ParsingContext<'source> {
                         let mut stmt = ctx.stmt_ctx();
                         let rest =
                             self.parse_expression(parser, ctx, &mut stmt, &mut continuing)?;
-                        ctx.lower(stmt, parser, rest, false, &mut continuing)?;
+                        ctx.lower(stmt, parser, rest, ExprPos::Rhs, &mut continuing)?;
                     }
                 }
 
-                let mut meta = meta.union(&self.expect(parser, TokenValue::RightParen)?.meta);
+                meta.subsume(self.expect(parser, TokenValue::RightParen)?.meta);
 
-                if let Some(stmt_meta) = self.parse_statement(parser, ctx, &mut block)? {
-                    meta = meta.union(&stmt_meta);
+                if let Some(stmt_meta) = self.parse_statement(parser, ctx, &mut block, &mut None)? {
+                    meta.subsume(stmt_meta);
                 }
 
                 body.push(
@@ -470,7 +505,7 @@ impl<'source> ParsingContext<'source> {
                         body: block,
                         continuing,
                     },
-                    meta.as_span(),
+                    meta,
                 );
 
                 ctx.remove_current_scope();
@@ -486,7 +521,8 @@ impl<'source> ParsingContext<'source> {
                 let meta = self.parse_compound_statement(meta, parser, ctx, &mut block)?;
 
                 ctx.remove_current_scope();
-                body.push(Statement::Block(block), meta.as_span());
+
+                body.push(Statement::Block(block), meta);
 
                 meta
             }
@@ -497,35 +533,43 @@ impl<'source> ParsingContext<'source> {
                 // turned into an error.
                 let mut stmt = ctx.stmt_ctx();
                 let expr = self.parse_expression(parser, ctx, &mut stmt, body)?;
-                ctx.lower(stmt, parser, expr, false, body)?;
+                ctx.lower(stmt, parser, expr, ExprPos::Rhs, body)?;
                 self.expect(parser, TokenValue::Semicolon)?.meta
             }
         };
 
-        Ok(Some(meta.union(&meta_rest)))
+        meta.subsume(meta_rest);
+        Ok(Some(meta))
     }
 
     pub fn parse_compound_statement(
         &mut self,
-        mut meta: SourceMetadata,
+        mut meta: Span,
         parser: &mut Parser,
         ctx: &mut Context,
         body: &mut Block,
-    ) -> Result<SourceMetadata> {
+    ) -> Result<Span> {
+        let mut terminator = None;
         loop {
             if let Some(Token {
                 meta: brace_meta, ..
             }) = self.bump_if(parser, TokenValue::RightBrace)
             {
-                meta = meta.union(&brace_meta);
+                meta.subsume(brace_meta);
                 break;
             }
 
-            if let Some(stmt_meta) = self.parse_statement(parser, ctx, body)? {
-                meta = meta.union(&stmt_meta);
+            let stmt = self.parse_statement(parser, ctx, body, &mut terminator)?;
+
+            if let Some(stmt_meta) = stmt {
+                meta.subsume(stmt_meta);
             }
         }
 
+        if let Some(idx) = terminator {
+            body.cull(idx..)
+        }
+
         Ok(meta)
     }
 
@@ -535,6 +579,10 @@ impl<'source> ParsingContext<'source> {
         context: &mut Context,
         body: &mut Block,
     ) -> Result<()> {
+        if self.bump_if(parser, TokenValue::Void).is_some() {
+            return Ok(());
+        }
+
         loop {
             if self.peek_type_name(parser) || self.peek_parameter_qualifier(parser) {
                 let qualifier = self.parse_parameter_qualifier(parser);
