@@ -1,12 +1,12 @@
 use super::{
     analyzer::{FunctionInfo, GlobalUse},
-    Capabilities, Disalignment, FunctionError, ModuleInfo, ShaderStages, TypeFlags,
-    ValidationFlags,
+    Capabilities, Disalignment, FunctionError, ModuleInfo,
 };
-use crate::arena::{Arena, Handle};
+use crate::arena::{Handle, UniqueArena};
 
 use bit_set::BitSet;
 
+#[cfg(feature = "validate")]
 const MAX_WORKGROUP_SIZE: u32 = 0x4000;
 
 #[derive(Clone, Debug, thiserror::Error)]
@@ -15,15 +15,10 @@ pub enum GlobalVariableError {
     InvalidUsage,
     #[error("Type isn't compatible with the storage class")]
     InvalidType,
-    #[error("Storage access {seen:?} exceeds the allowed {allowed:?}")]
-    InvalidStorageAccess {
-        allowed: crate::StorageAccess,
-        seen: crate::StorageAccess,
-    },
     #[error("Type flags {seen:?} do not meet the required {required:?}")]
     MissingTypeFlags {
-        required: TypeFlags,
-        seen: TypeFlags,
+        required: super::TypeFlags,
+        seen: super::TypeFlags,
     },
     #[error("Capability {0:?} is not supported")]
     UnsupportedCapability(Capabilities),
@@ -53,6 +48,8 @@ pub enum VaryingError {
     BindingCollision { location: u32 },
     #[error("Built-in {0:?} is present more than once")]
     DuplicateBuiltIn(crate::BuiltIn),
+    #[error("Capability {0:?} is not supported")]
+    UnsupportedCapability(Capabilities),
 }
 
 #[derive(Clone, Debug, thiserror::Error)]
@@ -81,6 +78,7 @@ pub enum EntryPointError {
     Function(#[from] FunctionError),
 }
 
+#[cfg(feature = "validate")]
 fn storage_usage(access: crate::StorageAccess) -> GlobalUse {
     let mut storage_usage = GlobalUse::QUERY;
     if access.contains(crate::StorageAccess::LOAD) {
@@ -96,9 +94,10 @@ struct VaryingContext<'a> {
     ty: Handle<crate::Type>,
     stage: crate::ShaderStage,
     output: bool,
-    types: &'a Arena<crate::Type>,
+    types: &'a UniqueArena<crate::Type>,
     location_mask: &'a mut BitSet,
     built_in_mask: u32,
+    capabilities: Capabilities,
 }
 
 impl VaryingContext<'_> {
@@ -160,6 +159,17 @@ impl VaryingContext<'_> {
                                 width,
                             },
                     ),
+                    Bi::ViewIndex => (
+                        match self.stage {
+                            St::Vertex | St::Fragment => !self.output,
+                            St::Compute => false,
+                        },
+                        *ty_inner
+                            == Ti::Scalar {
+                                kind: Sk::Sint,
+                                width,
+                            },
+                    ),
                     Bi::FragDepth => (
                         self.stage == St::Fragment && self.output,
                         *ty_inner
@@ -176,6 +186,21 @@ impl VaryingContext<'_> {
                                 width: crate::BOOL_WIDTH,
                             },
                     ),
+                    Bi::PrimitiveIndex => {
+                        if !self.capabilities.contains(Capabilities::PRIMITIVE_INDEX) {
+                            return Err(VaryingError::UnsupportedCapability(
+                                Capabilities::PRIMITIVE_INDEX,
+                            ));
+                        }
+                        (
+                            self.stage == St::Fragment && !self.output,
+                            *ty_inner
+                                == Ti::Scalar {
+                                    kind: Sk::Uint,
+                                    width,
+                                },
+                        )
+                    }
                     Bi::SampleIndex => (
                         self.stage == St::Fragment && !self.output,
                         *ty_inner
@@ -203,7 +228,8 @@ impl VaryingContext<'_> {
                     Bi::GlobalInvocationId
                     | Bi::LocalInvocationId
                     | Bi::WorkGroupId
-                    | Bi::WorkGroupSize => (
+                    | Bi::WorkGroupSize
+                    | Bi::NumWorkGroups => (
                         self.stage == St::Compute && !self.output,
                         *ty_inner
                             == Ti::Vector {
@@ -231,10 +257,6 @@ impl VaryingContext<'_> {
                     return Err(VaryingError::BindingCollision { location });
                 }
 
-                // Values passed from the vertex shader to the fragment shader must have their
-                // interpolation defaulted (i.e. not `None`) by the front end, as appropriate for
-                // that language. For anything other than floating-point scalars and vectors, the
-                // interpolation must be `Flat`.
                 let needs_interpolation = match self.stage {
                     crate::ShaderStage::Vertex => self.output,
                     crate::ShaderStage::Fragment => !self.output,
@@ -296,37 +318,39 @@ impl VaryingContext<'_> {
 }
 
 impl super::Validator {
+    #[cfg(feature = "validate")]
     pub(super) fn validate_global_var(
         &self,
         var: &crate::GlobalVariable,
-        types: &Arena<crate::Type>,
+        types: &UniqueArena<crate::Type>,
     ) -> Result<(), GlobalVariableError> {
+        use super::TypeFlags;
+
         log::debug!("var {:?}", var);
         let type_info = &self.types[var.ty.index()];
 
-        let (allowed_storage_access, required_type_flags, is_resource) = match var.class {
+        let (required_type_flags, is_resource) = match var.class {
             crate::StorageClass::Function => return Err(GlobalVariableError::InvalidUsage),
-            crate::StorageClass::Storage => {
+            crate::StorageClass::Storage { .. } => {
                 if let Err((ty_handle, disalignment)) = type_info.storage_layout {
-                    if self.flags.contains(ValidationFlags::STRUCT_LAYOUTS) {
+                    if self.flags.contains(super::ValidationFlags::STRUCT_LAYOUTS) {
                         return Err(GlobalVariableError::Alignment(ty_handle, disalignment));
                     }
                 }
                 (
-                    crate::StorageAccess::all(),
                     TypeFlags::DATA | TypeFlags::HOST_SHARED | TypeFlags::TOP_LEVEL,
                     true,
                 )
             }
             crate::StorageClass::Uniform => {
                 if let Err((ty_handle, disalignment)) = type_info.uniform_layout {
-                    if self.flags.contains(ValidationFlags::STRUCT_LAYOUTS) {
+                    if self.flags.contains(super::ValidationFlags::STRUCT_LAYOUTS) {
                         return Err(GlobalVariableError::Alignment(ty_handle, disalignment));
                     }
                 }
                 (
-                    crate::StorageAccess::empty(),
                     TypeFlags::DATA
+                        | TypeFlags::COPY
                         | TypeFlags::SIZED
                         | TypeFlags::HOST_SHARED
                         | TypeFlags::TOP_LEVEL,
@@ -334,23 +358,15 @@ impl super::Validator {
                 )
             }
             crate::StorageClass::Handle => {
-                let access = match types[var.ty].inner {
-                    crate::TypeInner::Image {
-                        class: crate::ImageClass::Storage(_),
-                        ..
-                    } => crate::StorageAccess::all(),
-                    crate::TypeInner::Image { .. } | crate::TypeInner::Sampler { .. } => {
-                        crate::StorageAccess::empty()
-                    }
+                match types[var.ty].inner {
+                    crate::TypeInner::Image { .. } | crate::TypeInner::Sampler { .. } => {}
                     _ => return Err(GlobalVariableError::InvalidType),
                 };
-                (access, TypeFlags::empty(), true)
+                (TypeFlags::empty(), true)
+            }
+            crate::StorageClass::Private | crate::StorageClass::WorkGroup => {
+                (TypeFlags::DATA | TypeFlags::SIZED, false)
             }
-            crate::StorageClass::Private | crate::StorageClass::WorkGroup => (
-                crate::StorageAccess::empty(),
-                TypeFlags::DATA | TypeFlags::SIZED,
-                false,
-            ),
             crate::StorageClass::PushConstant => {
                 if !self.capabilities.contains(Capabilities::PUSH_CONSTANT) {
                     return Err(GlobalVariableError::UnsupportedCapability(
@@ -358,20 +374,12 @@ impl super::Validator {
                     ));
                 }
                 (
-                    crate::StorageAccess::LOAD,
-                    TypeFlags::DATA | TypeFlags::HOST_SHARED | TypeFlags::SIZED,
+                    TypeFlags::DATA | TypeFlags::COPY | TypeFlags::HOST_SHARED | TypeFlags::SIZED,
                     false,
                 )
             }
         };
 
-        if !allowed_storage_access.contains(var.storage_access) {
-            return Err(GlobalVariableError::InvalidStorageAccess {
-                seen: var.storage_access,
-                allowed: allowed_storage_access,
-            });
-        }
-
         if !type_info.flags.contains(required_type_flags) {
             return Err(GlobalVariableError::MissingTypeFlags {
                 seen: type_info.flags,
@@ -392,9 +400,12 @@ impl super::Validator {
         module: &crate::Module,
         mod_info: &ModuleInfo,
     ) -> Result<FunctionInfo, EntryPointError> {
+        #[cfg(feature = "validate")]
         if ep.early_depth_test.is_some() && ep.stage != crate::ShaderStage::Fragment {
             return Err(EntryPointError::UnexpectedEarlyDepthTest);
         }
+
+        #[cfg(feature = "validate")]
         if ep.stage == crate::ShaderStage::Compute {
             if ep
                 .workgroup_size
@@ -407,16 +418,21 @@ impl super::Validator {
             return Err(EntryPointError::UnexpectedWorkgroupSize);
         }
 
-        let stage_bit = match ep.stage {
-            crate::ShaderStage::Vertex => ShaderStages::VERTEX,
-            crate::ShaderStage::Fragment => ShaderStages::FRAGMENT,
-            crate::ShaderStage::Compute => ShaderStages::COMPUTE,
-        };
-
         let info = self.validate_function(&ep.function, module, mod_info)?;
 
-        if !info.available_stages.contains(stage_bit) {
-            return Err(EntryPointError::ForbiddenStageOperations);
+        #[cfg(feature = "validate")]
+        {
+            use super::ShaderStages;
+
+            let stage_bit = match ep.stage {
+                crate::ShaderStage::Vertex => ShaderStages::VERTEX,
+                crate::ShaderStage::Fragment => ShaderStages::FRAGMENT,
+                crate::ShaderStage::Compute => ShaderStages::COMPUTE,
+            };
+
+            if !info.available_stages.contains(stage_bit) {
+                return Err(EntryPointError::ForbiddenStageOperations);
+            }
         }
 
         self.location_mask.clear();
@@ -429,6 +445,7 @@ impl super::Validator {
                 types: &module.types,
                 location_mask: &mut self.location_mask,
                 built_in_mask: argument_built_ins,
+                capabilities: self.capabilities,
             };
             ctx.validate(fa.binding.as_ref())
                 .map_err(|e| EntryPointError::Argument(index as u32, e))?;
@@ -444,6 +461,7 @@ impl super::Validator {
                 types: &module.types,
                 location_mask: &mut self.location_mask,
                 built_in_mask: 0,
+                capabilities: self.capabilities,
             };
             ctx.validate(fr.binding.as_ref())
                 .map_err(EntryPointError::Result)?;
@@ -452,6 +470,8 @@ impl super::Validator {
         for bg in self.bind_group_masks.iter_mut() {
             bg.clear();
         }
+
+        #[cfg(feature = "validate")]
         for (var_handle, var) in module.global_variables.iter() {
             let usage = info[var_handle];
             if usage.is_empty() {
@@ -461,12 +481,12 @@ impl super::Validator {
             let allowed_usage = match var.class {
                 crate::StorageClass::Function => unreachable!(),
                 crate::StorageClass::Uniform => GlobalUse::READ | GlobalUse::QUERY,
-                crate::StorageClass::Storage => storage_usage(var.storage_access),
+                crate::StorageClass::Storage { access } => storage_usage(access),
                 crate::StorageClass::Handle => match module.types[var.ty].inner {
                     crate::TypeInner::Image {
-                        class: crate::ImageClass::Storage(_),
+                        class: crate::ImageClass::Storage { access, .. },
                         ..
-                    } => storage_usage(var.storage_access),
+                    } => storage_usage(access),
                     _ => GlobalUse::READ | GlobalUse::QUERY,
                 },
                 crate::StorageClass::Private | crate::StorageClass::WorkGroup => GlobalUse::all(),
